import os
import sys
import openpyxl
from math import exp, sqrt
import numpy as np
from math import log, exp
from numpy import arange
from scipy import integrate         # 数値積分関数 integrateを読み込む
from scipy import optimize          # newton関数はscipy.optimizeモジュールに入っている
from scipy.optimize import minimize
from scipy.interpolate import interp1d
from matplotlib import pyplot as plt


from tklib.tkfile import tkFile
from tklib.tkinifile import tkIniFile
import tklib.tkre as tkre
from tklib.tkutils import mprint, IsDir, IsFile, SplitFilePath, modify_path, delete_file
from tklib.tkutils import terminate, save_csv, safe_getelement, validate_error
from tklib.tkutils import pint, pfloat, getarg, getintarg, getfloatarg
from tklib.tksci.tkmatrix import make_matrix1, make_matrix2, make_matrix3
from tklib.tksci.tksci import Reduce01, Round
from tklib.tksci.tksci import pi, h, hbar, e, c, e0, kB, me, RyToeV, HartreeToeV, JToeV, eVToJ
from tklib.tksci.tkmatrix import make_matrix1, make_matrix2, make_matrix3
from tklib.tksci.tksmoothing import simple_average, polynomial_fit
from tklib.tkcrystal.tkcif import tkCIF, tkCIFData
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tksci import tkequation
import tklib.tkcsv as tkcsv
from tklib.tkexcel import tkExcel

from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tkcrystal.tkboltztrap import tkBoltzTraP
from tklib.tktransport.tkTransport import fe, fh
from tklib.tktransport.tkTransport import integrate_Simpson_list, fe, fh, NC2meff_FEA, meff2DC0_FEA, meff2NC_FEA
from tklib.tktransport.tkDOS_VASP import integrate_Simpson_list, tkDOS


"""
BoltzTraP2の計算結果から、輸送物性の温度依存性を二分法で計算。
"""

ekB = e / kB

vasp = tkVASP()
BT   = tkBoltzTraP()

# mode: 'S', 'sigma', 'kappa', 'PF', 'ZT', 'conv', 'NC'
mode = 'S'
#mode = 'sigma'
#mode = 'conv'

# convert mode: 'T', 'tensor', 'all'
#conv_mode = 'T'
conv_mode = 'tensor'
#conv_mode = 'all'

# input file and dir
infile          = '.\case'
Work_dir        = '.'
parameterfile   = None
parameterbkfile = None

# Interpolation type
interp_type = 'linear'
#interp_type = 'cubic'

# tau0 and lattice thermal conductivity
tau0       = 1.0e-14    # s
kappa_latt = 2.0        # W/m/K

# EF range to read
EFmin = -10.0 # eV
EFmax =  10.0

# integration parameters
# Integration range w.r.t. kBT
nrange  = 10.0

# # of smoothing point to interporate tensors at estimated EF(T)
nsmooth = 1

# Plot EF range
dEFmin = -0.5  # measured from EV, eV
dEFmax =  0.5  # measured from EC, eV
nEF    = 201
EFstep = None

dos = tkDOS()
dos.data_for_bandedges = 'DOSCAR'

dos.T0 = 300.0      # Temperature K
dos.dEA  =  0.05    # Acceptor level measured from EV, eV
dos.NA   =  0.0e17  # cm-3
dos.dED  = -0.05    # Donor level measured from EC, eV
dos.ND   =  0.0e17  # cm-3

dos.EFatT0 = 10.0      # EF at T0 [K]. If abs(EFatT0) >= 5.0, this parameter will not be considered
# Fermi level from EF(VASP) at 0 K
dEF0 = 0.0 # eV

dos.Egth =  1.0e18  # Critera DOS value to search band edges, cm^-3

dos.interp_type = 'linear'
#dos.interp_type = 'cubic'
# Integration range w.r.t. kBT
dos.Einteg0 = -3.0
dos.Einteg1 =  3.0
dos.nrange  =  6.0
# Relative accuracy of the quad functin
dos.eps_quad     = 1.0e-5
dos.nmaxdiv_quad = 150

# Bisection method parameters
# Initial search range
dos.dE_bisec    = 2.0
# EPS
dos.eps_bisec   = 1.0e-7
# max iteration number
dos.nmaxiter_bisec      = 200
# Output cycle
dos.iprintiterval_bisec = 1


E_raw   = None
dos_raw = None
EDOSmin  = None
EDOSmax  = None
EDOSstep = None
fdos = None


#=============================
# Graph configuration
#=============================
view_Emin = EFmin # eV
view_Emax = EFmax
nT_view = 5

fig = None
figsize = [14, 10]
fontsize = 16
legend_fontsize = 8
graphupdateinterval = 10


#=============================
# other functions
#=============================
def usage():
    global Work_dir
    
    if Work_dir == '':
        Work_dir = '.'
        
    print("(i) usage: python {} NC BoltzTraP_input_file T nsmooth".format(sys.argv[0]))
    print("      Calculate NC and NV at given T")
    print("  BoltzTraP_input_file: Specify outputtrans or transdos file, or DOSCAR path")
    print("  nsmooth: # of smooting points for DOS".format(nsmooth))
    print("   ex: python {} {} {} {} {}".format(sys.argv[0], 'NC', infile, dos.T0, nsmooth))
    print("(ii) usage: python {} [S|sigma|kappa] BoltzTraP_input_file tau0 kappa_latt dEF0 T0 EF_at_T0 nsmooth".format(sys.argv[0]))
#    print("(ii) usage: python {} [S|sigma|kappa|PF|ZT] dir tau0 kappa_latt dEF0 EF0K nsmooth".format(sys.argv[0]))
    print("      Calculate T dependences of the property given by mode")
#    print("  mode: 'S', 'sigma', 'kappa'");
    print("  mode: 'S', 'sigma', 'kappa', 'PF', 'ZT'");
    print("  kappa_latt: lattice thermal conductivity in W/m/K")
    print("    dEF0: Fermi level shift from EF(VASP) to search band edge levels")
    print("    EF0K: Fermi level at 0K. Used to change the doping level")
    print("   ex: python {} {} {} {} {} {} {} {}".format(sys.argv[0], 'S',     infile, tau0, kappa_latt, dEF0, dos.T0, dos.EFatT0, nsmooth))
    print("   ex: python {} {} {} {} {} {} {} {}".format(sys.argv[0], 'sigma', infile, tau0, kappa_latt, dEF0, dos.T0, dos.EFatT0, nsmooth))
    print("   ex: python {} {} {} {} {} {} {} {}".format(sys.argv[0], 'kappa', infile, tau0, kappa_latt, dEF0, dos.T0, dos.EFatT0, nsmooth))
    print("   ex: python {} {} {} {} {} {} {} {}".format(sys.argv[0], 'PF',    infile, tau0, kappa_latt, dEF0, dos.T0, dos.EFatT0, nsmooth))
    print("   ex: python {} {} {} {} {} {} {} {}".format(sys.argv[0], 'ZT',    infile, tau0, kappa_latt, dEF0, dos.T0, dos.EFatT0, nsmooth))
    print("(iii) usage: python {} conv conv_mode BoltzTraP_input_file tau0 kappa_latt".format(sys.argv[0]))
    print("  conv_mode: 'T', 'tensor', 'all'");
    print("      Specify which data are included in each file")
    print("  kappa_latt: lattice thermal conductivity in W/m/K")
    print("   ex: python {} {} {} {} {} {}".format(sys.argv[0], 'conv', conv_mode, infile, tau0, kappa_latt))

def updatevars():
    global mode, conv_mode
    global infile, Work_dir, parameterfile, parameterbkfile
    global dEF0, nsmooth
    global T0, tau0, kappa_latt

    argv = sys.argv
#    if len(argv) <= 1:
#        terminate()

    mode    = getarg(     1, mode)
    if mode == 'NC':
        infile = getarg(     2, infile)
        dos.T0  = getfloatarg(3, dos.T0)
        nsmooth = getintarg  (4, nsmooth)
    elif mode == 'S' or mode == 'sigma' or mode == 'kappa' or mode == 'PF' or mode == 'ZT':
        infile    = getarg(     2, infile)
        tau0       = getfloatarg(3, tau0)
        kappa_latt = getfloatarg(4, kappa_latt)
        dEF0       = getfloatarg(5, dEF0)
        dos.T0     = getfloatarg(6, dos.T0)
        dos.EFatT0 = getfloatarg(7, dos.EFatT0)
        nsmooth    = getintarg  (8, nsmooth)
    elif mode == 'conv':
        conv_mode   = getarg(     2, conv_mode)
        infile     = getarg(     3, infile)
        tau0        = getfloatarg(4, tau0)
        kappa_latt  = getfloatarg(5, kappa_latt)
    else:
        terminate("Error in updatevars(): Invalid mode [{}]".format(mode), usage = usage)

    header, ext     = os.path.splitext(infile)
    filebody        = os.path.basename(header)
    parameterfile   = filebody + '.prm'
    parameterbkfile = filebody + '.prm.bak'

def get_parameters(key = None):
    if key is None:
        return {}
    elif key == 'Preferences':
        return {
            "nrange": nrange, 
            "Einteg0": dos.Einteg0, "Einteg1": dos.Einteg1, "nrange": nrange,
            "dE_bisec": dos.dE_bisec,
            "nsmooth": nsmooth
            }
    elif key == 'ElectronicStructure':
        return {
            "tau0": tau0, "kappa_latt": kappa_latt, 
            "EV": dos.EV, "EC": dos.EC, 
            "dEA": dos.dEA, "NA": dos.NA, 
            "dED": dos.dED, "ND": dos.ND,
            "T0": dos.T0, "dEF0": dEF0, "Egth": dos.Egth, "EFatT0": dos.EFatT0
            }
    elif key == 'Graph':
        return {
            "dEFmin": dEFmin, "dEFmax": dEFmax, "nEF": nEF,
            "EFmin": EFmin, "EFmax": EFmax,
            "view_Emin": view_Emin, "view_Emax": view_Emax, "nT_view": nT_view,
            "figsize_x": figsize[0], "figsize_y": figsize[1],
            "fontsize": fontsize, "legend_fontsize": legend_fontsize
            }

def save_parameters(path, varname = None, ai = None, optid = None, section = 'Parameters'):
    if varname is None:
        varname = get_parameters()
    
    ini = tkIniFile(path)
    if type(varname) is list:
        if optid is None:
            for i in range(len(varname)):
                ini.WriteString(section, varname[i], ai[i])
        else:
            for i in range(len(varname)):
                ini.WriteString(section, varname[i], "{}:{}".format(ai[i], optid[i]))
    else:
        for key in varname:
            ini.WriteString(section, key, "{}".format(varname[key]))

def save_parameterfile(parameterfile):
    save_parameters(parameterfile, get_parameters("Preferences"), section = 'Preferences')
    save_parameters(parameterfile, get_parameters("ElectronicStructure"), section = 'ElectronicStructure')
    save_parameters(parameterfile, get_parameters("Graph"), section = 'Graph')

def read_parameters(path, varname = None, ai = None, optid = None, update_params = 0, AddSection = 0, repeat_input = 0):
    global nrange, nsmooth
    global tau0, kappa_latt, dEF0
    global dEFmin, dEFmax, nEF, EFmin, EFmax, view_Emin, view_Emax, nT_view
    global figsize, fontsize, legend_fontsize

    ini = tkIniFile()
    inf = ini.ReadAll(path, AddSection = 0)
    if inf is None:
        return None

    nrange       = pfloat(safe_getelement(inf, 'nrange', nrange))
    nsmooth      = pint(safe_getelement(inf, 'nsmooth', nsmooth))
    dos.Einteg0  = pfloat(safe_getelement(inf, 'Einteg0', dos.Einteg0))
    dos.Einteg1  = pfloat(safe_getelement(inf, 'Einteg1', dos.Einteg1))
    dos.dE_bisec = pfloat(safe_getelement(inf, 'dE_bisec', dos.dE_bisec))

    tau0 = pfloat(safe_getelement(inf, 'tau0', tau0))
    kappa_latt = pfloat(safe_getelement(inf, 'kappa_latt', kappa_latt))
    dos.EV     = pfloat(safe_getelement(inf, 'EV',     dos.EV))
    dos.EC     = pfloat(safe_getelement(inf, 'EC',     dos.EC))
    dos.dEA    = pfloat(safe_getelement(inf, 'dEA',    dos.dEA))
    dos.NA     = pfloat(safe_getelement(inf, 'NA',     dos.NA))
    dos.dED    = pfloat(safe_getelement(inf, 'dED',    dos.dED))
    dos.ND     = pfloat(safe_getelement(inf, 'ND',     dos.ND))
    dos.T0     = pfloat(safe_getelement(inf, 'T0',     dos.T0))
    dEF0       = pfloat(safe_getelement(inf, 'dEF0',   dEF0))
    dos.Egth   = pfloat(safe_getelement(inf, 'Egth',   dos.Egth))
    dos.EFatT0 = pfloat(safe_getelement(inf, 'EFatT0', dos.EFatT0))

    dEFmin     = pfloat(safe_getelement(inf, 'dEFmin', dEFmin))
    dEFmax     = pfloat(safe_getelement(inf, 'dEFmax', dEFmax))
    nEF        = pint  (safe_getelement(inf, 'nEF',    nEF))
    EFmin      = pfloat(safe_getelement(inf, 'EFmin',  EFmin))
    EFmax      = pfloat(safe_getelement(inf, 'EFmax',  EFmax))
    view_Emin  = pfloat(safe_getelement(inf, 'view_Emin', view_Emin))
    view_Emax  = pfloat(safe_getelement(inf, 'view_Emax', view_Emax))
    nT_view    = pint  (safe_getelement(inf, 'nT_view',   nT_view))

    figsize[0]      = pint(safe_getelement(inf, 'figsize_x', figsize[0]))
    figsize[1]      = pint(safe_getelement(inf, 'figsize_y', figsize[1]))
    fontsize        = pint(safe_getelement(inf, 'fontsize', fontsize))
    legend_fontsize = pint(safe_getelement(inf, 'legend_fontsize', legend_fontsize))

    return inf

def print_parameters():
    print("Preferences")
    vars = get_parameters("Preferences")
    for key in vars.keys():
        if vars[key] is None:
            print("  {:10s}: {:14}".format(key, 'None'))
        else:
            print("  {:10s}: {:14.8g}".format(key, vars[key]))
    print("Electronic Structure")
    vars = get_parameters("ElectronicStructure")
    for key in vars.keys():
        if vars[key] is None:
            print("  {:10s}: {:14}".format(key, 'None'))
        else:
            print("  {:10s}: {:14.8g}".format(key, vars[key]))
    print("Graph")
    vars = get_parameters("Graph")
    for key in vars.keys():
        if vars[key] is None:
            print("  {:10s}: {:14}".format(key, 'None'))
        else:
            print("  {:10s}: {:14.8g}".format(key, vars[key]))

def average_dos(E, y, nsmooth, EV, EC):
    m = int(nsmooth / 2);
    nsmooth = m + m +1

    Estart = E[0]
    Estep  = E[1] - E[0]
    iEV = int((EV - Estart) / Estep + Estep * 0.5)
    iEC = int((EC - Estart) / Estep + Estep * 0.5)

    if iEV == iEC:
        return simple_average(y, nsmooth)

#    ysVB = simple_average(y[0:iEV+1], nsmooth)
#    ysEG = simple_average(y[iEV+1:iEC], nsmooth)
#    ysCB = simple_average(y[iEC:], nsmooth)
    ysVB = polynomial_fit(y[0:iEV+1], nsmooth)
    ysEG = polynomial_fit(y[iEV+1:iEC], nsmooth)
    ysCB = polynomial_fit(y[iEC:], nsmooth)

    ys = ysVB + ysEG + ysCB
    
    return ys

def read_files(base_path, normalize_E = True):
    global nEF, EFmin, EFmax, EFstep

    allinf = {}

    Work_dir, basename, filebody, ext = SplitFilePath(base_path)
    if Work_dir == '':
        Work_dir = '.'
    DOSCAR_path     = vasp.get_DOSCAR(Work_dir)
    intranspath     = BT.get_intranspath(filebody, Work_dir)
    outputtranspath = BT.get_outputtranspath(filebody, Work_dir)
    transdospath    = BT.get_transdospath(filebody, Work_dir)
    condtenspath    = BT.get_condtenspath(filebody, Work_dir)
    halltenspath    = BT.get_halltenspath(filebody, Work_dir)
    tracepath       = BT.get_tracepath(filebody, Work_dir)
#    header, ext     = os.path.splitext(DOSCAR_path)
#    filebody        = os.path.basename(header)
#    outcsvfile      = filebody + f'-EFT-{mode}.csv'
    outcsvfile      = os.path.join(Work_dir, filebody + f'-EFT-{mode}.csv')
    
    allinf["transdospath"] = transdospath
    allinf["condtenspath"] = condtenspath
    allinf["halltenspath"] = halltenspath
    allinf["tracepath"]    = tracepath
    allinf["outcsvfile"]   = outcsvfile

    print("mode:", mode)
    print("Work dir   : {}".format(Work_dir))
    print("DOSCAR     : {}".format(DOSCAR_path))
    print("intrans    : {}".format(intranspath))
    print("outputtrans: {}".format(outputtranspath))
    print("transdos   : {}".format(transdospath))
#    print("condtens   : {}".format(condtenspath))
    print("trace      : {}".format(tracepath))
    print("Output CSV : {}".format(outcsvfile))

    print("")
    print("Read VASP files from {}".format(Work_dir))
    if normalize_E:
        dos.read_vaspfiles(Work_dir, EF0 = 0.0, normalize_E = normalize_E, unit = '/cm3', data_for_bandedges = 'EIGENVAL', 
                print_level = 1, exit_by_error = False)
    else:
        dos.read_vaspfiles(Work_dir, EF0 = None, normalize_E = normalize_E, unit = '/cm3', data_for_bandedges = 'EIGENVAL', 
                print_level = 1, exit_by_error = False)

    outcarinf   = dos.vaspinf["OUTCAR"]
    ISPIN       = outcarinf.get("ISPIN", 1)
    if ISPIN == 1:
        SpinPolarized = False
    else:
        SpinPolarized = True
    cry         = dos.vaspinf["POSCAR"]
    EF_vasp0    = outcarinf["EF"]

    print("")
    print("Read intrans from {}".format(intranspath))
    inf_inbt = BT.read_intrans(intranspath, exit_by_error = False)
    if inf_inbt is None:
        print("  Warning: {} does not exist.".format(intranspath))
        inf_inbt = {}
    else:
        print("  inf:", inf_inbt)
    allinf["inf_intrans"] = inf_inbt

    print("")
    print("Read outputtrans from {}".format(outputtranspath))
    inf_outbt = BT.read_outputtrans(outputtranspath, exit_by_error = False)
    if inf_outbt is None:
        print("  Warning: {} does not exist.".format(outputtranspath))
        inf_outbt = {}
        EF_bt0 = EF_vasp0
        print("           EF0 is taken from VASP OUTCAR EF0={} eV".format(EF_bt0))
    else:
        print("  inf:", inf_outbt)
        allinf["inf_outputtrans"] = inf_outbt
        EF_bt0 = inf_outbt["FermiE"] * RyToeV
        print("   EF_bt0: {} eV".format(EF_bt0))
        print("   EVBM0 : {} eV".format(inf_outbt["EVBM"]))
        print("   ECBM0 : {} eV".format(inf_outbt["ECBM"]))
        if normalize_E:
            inf_outbt["EVBM"] = inf_outbt["EVBM"] * RyToeV - EF_bt0
            inf_outbt["ECBM"] = inf_outbt["ECBM"] * RyToeV - EF_bt0
               
    print("")
    print("Read trace from {}".format(tracepath))
    inf = BT.read_trace(tracepath, dos.Vcell, E0 = EF_bt0, normalize_E= normalize_E, unit = '/cm3', EFmin = None, EFmax = None, 
#    inf = BT.read_trace(tracepath, dos.Vcell, E0 = EF_vasp0, normalize_E= normalize_E, unit = '/cm3', EFmin = None, EFmax = None, 
                print_level = 0, exit_by_error = False)
    allinf["inf_trace"] = inf

    print("")
    print("Read transdos from {}".format(transdospath))
    inf_dosbt = BT.read_transdos(transdospath, dos.Vcell, E0 = EF_bt0, normalize_E= normalize_E, unit = '/cm3', unit_N = '', 
                        SpinPolarized = SpinPolarized, Emin = None, Emax = None, 
                print_level = 0, exit_by_error = False)
    allinf["inf_trancedos"] = inf_dosbt

    return allinf

def cal_NC():
    print("")
    print("===============================================================================")
    print("  Calculate NC and NV from .transdos of BoltzTraP2")
    print("===============================================================================")

    allinf     = read_files(infile)
    inf_inbt   = allinf.get("inf_intrans", None)
    inf_outbt  = allinf.get("inf_outputtrans", None)
    inf_trace  = allinf.get("inf_trace", None)
    outcsvfile = allinf.get("outcsvfile", None)
    outcarinf  = dos.vaspinf.get("OUTCAR", None)
    ISPIN      = outcarinf.get("ISPIN", 1)
    if ISPIN == 1:
        SpinPolarized = False
    else:
        SpinPolarized = True
    cry        = dos.vaspinf["POSCAR"]
    EF_vasp0   = outcarinf["EF"]
    EF_vasp    = dos.EF # outcarinf["EF"]
    dos.EF0    = EF_vasp + dEF0
    EFmin      = dos.EV + dEFmin
    EFmax      = dos.EC + dEFmax
    EFstep     = (EFmax - EFmin) / (nEF - 1)
    E_vasp     = dos.Elist
    EV_vasp    = dos.EV
    EC_vasp    = dos.EC
    DOS_vasp   = dos.DOSlist
    print("")
    print("VASP info")
    print("  ISPIN        : {}".format(ISPIN))
    print("  EF(VASP)(org): {} eV".format(EF_vasp0))
    print("  EF(VASP)     : {} eV".format(EF_vasp))
    print("  EF0          : {} eV".format(dos.EF0))
    print("  EV           : {} eV".format(dos.EV))
    print("  EC           : {} eV".format(dos.EC))
    print("  Vcell        : {} A^-3".format(dos.Vcell))
    print("  DOS E range  : {:8.3g} - {:8.3g} eV".format(E_vasp[0], E_vasp[-1]))
    print("  Plot EF range: {} - {} eV at {} eV step for {} points".format(EFmin, EFmax, EFstep, nEF))

    dos.outrangeE = None
    DOS_source  = 'DOSCAR'
    EV    = dos.EV
    EC    = dos.EC
    Eg    = dos.Eg
    EHOMO = dos.EHOMO
    ELUMO = dos.ELUMO
    print("  Band edges: EV={:10.6f}  EC={:10.6f}  Eg={:10.6f}  EF(VASP)={:10.6f} eV"
        .format(EV, EC, Eg, EF_vasp))

    if inf_outbt is not None:
        print("")
        print("BoltzTraP2 info")
#        EVBM_bt = inf_outbt["EVBM"] + EF_vasp0
#        ECBM_bt = inf_outbt["ECBM"] + EF_vasp0
        print(".outputtrans exists but EV and EC are taken from VASP")
        EVBM_bt = EV_vasp
        ECBM_bt = EC_vasp
    else:
        print("")
        print(".outputtrans does not exist. EV and EC are taken from VASP")
        EVBM_bt = EV_vasp
        ECBM_bt = EC_vasp
    print("  EV           : {} eV".format(EVBM_bt))
    print("  EC           : {} eV".format(ECBM_bt))

    if inf_trace is None:
        E_DOS_bt     = None
        DOS_bt       = None
        DOS_bt_label = None
    else:
        E_DOS_bt     = inf_trace['xEF']
        DOS_bt       = inf_trace['dos'][0][0]
        DOS_bt_label = inf_trace['dos_label']
        DOS_source  = 'trace'
#        dos.set_dos(E_DOS_bt, DOS_bt, EVBM = EVBM_bt, ECBM = ECBM_bt)

    inf_dosbt = allinf["inf_trancedos"]
    if inf_dosbt is None:
        nE         = 0
        E_DOS_trbt = None
        DOS_trbt   = None
        N_trbt     = None
    else:
        nE         = inf_dosbt["nE"]
        E_DOS_trbt = inf_dosbt["xE"]
        DOS_trbt   = inf_dosbt["dos"]
#        DOS_trbt   = simple_average(inf_dosbt["dos"], nsmooth)
#        DOS_trbt   = average_dos(E_DOS_trbt, inf_dosbt["dos"], nsmooth, dos.EV, dos.EC)
        N_trbt     = inf_dosbt["yNcm3"]
        DOS_source  = 'transdos'
#        dos.set_dos(E_DOS_trbt, DOS_trbt, EVBM = EVBM_bt, ECBM = ECBM_bt)

    print("")
    print(f"Use {DOS_source} DOS to calculate NC and NV")
    if DOS_source == 'transdos':
        DOS_trbt = average_dos(E_DOS_trbt, DOS_trbt, nsmooth, EVBM_bt, ECBM_bt)
        dos.set_dos(E_DOS_trbt, DOS_trbt, EVBM = EVBM_bt, ECBM = ECBM_bt)
    elif DOS_source == 'trace':
        DOS_bt   = average_dos(E_DOS_bt, DOS_bt, nsmooth, EVBM_bt, ECBM_bt)
        dos.set_dos(E_DOS_bt, DOS_bt, EVBM = EVBM_bt, ECBM = ECBM_bt)

    print("")
    print("Integration E range: {} - {} eV, or EF+-{}*kBT eV".format(dos.Einteg0, dos.Einteg1, dos.nrange))
    print("  quad() parameters")
    print("    releps=",  dos.eps_quad)
    print("    nmaxdiv=", dos.nmaxdiv_quad)

    print("")
    print("EF dependence")
    print("  EF range: {} - {}, {} eV step for {} points".format(EFmin, EFmax, EFstep, nEF))
    xEF = [EFmin + iEF * EFstep for iEF in range(nEF)]
    yNh, yNe, yNAm, yNDp, yRH, yNs, yNsabs = dos.cal_EF_lists(dos.T0, xEF, print_level = 0)

    yT0Ne = []
    yT0Nh = []
    print("")
    print ("%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s\t%10s" % ("EF(eV)", "Ne", "Nh", "ND+", "NA-", "RH", "T0(Ne)", "T0(Nh)", "Ns"))
    for i in range(nEF):
        EF = xEF[i]
        
        if i == 0:
            i1 = i
            i2 = i+1
        elif i == nEF - 1:
            i1 = i-1
            i2 = i
        else:
            i1 = i-1
            i2 = i+1

        dEF = xEF[i2] - xEF[i1]
        dlnNedEF = (log(yNe[i2]+1.0e-300) - log(yNe[i1]+1.0e-300)) / dEF
        dlnNhdEF = (log(yNh[i2]+1.0e-300) - log(yNh[i1]+1.0e-300)) / dEF
        if dlnNedEF == 0.0:
            dlnNedEF = 1.0-4
        T0Ne =  1.0 / kB / dlnNedEF * e
        if dlnNhdEF == 0.0:
            dlnNhdEF = 1.0-4
        T0Nh = -1.0 / kB / dlnNhdEF * e
        yT0Ne.append(T0Ne)
        yT0Nh.append(T0Nh)

        if i % 10 == 0:
            print ("%10.4f\t%10.4g\t%10.4g\t%10.4g\t%10.4g\t%10.4g\t%10.4g\t%10.4g\t%10.4g" 
                    % (EF, yNe[i], yNh[i], yNDp[i], yNAm[i], yRH[i], T0Ne, T0Nh, yNs[i]))

    print("")
    Emid = (dos.EV + dos.EC) / 2.0
    imid = int((Emid - EFmin) / EFstep + 1.0e-5)
    Eimid = EFmin + imid * EFstep
    Ne   = yNe[imid]
    Nh   = yNh[imid]
    T0Ne = yT0Ne[imid]
    T0Nh = yT0Ne[imid]

    NC = Ne * exp((dos.EC - Eimid) * e / kB / T0Ne)
    NV = Nh * exp((Eimid - dos.EV) * e / kB / T0Nh)
    meeff = NC2meff_FEA(NC, dos.T0)
    mheff = NC2meff_FEA(NV, dos.T0)
    DC0   = meff2DC0_FEA(meeff, dos.T0)
    DV0   = meff2DC0_FEA(mheff, dos.T0)
    print("")
    print("Effective density of states at the mid gap {} eV (i={}, E={:8.4g}) by [{}]:".format(Emid, imid, Eimid, DOS_source))
    if DOS_source != 'transdos':
        print("    Not dens DOS. do not trust these values")
    print(f"  Source: {DOS_source}")
    print("   NC  = {:12.6g} cm^-3         (T0 = {:8.4g} K)".format(NC, T0Ne))
    print("   NV  = {:12.6g} cm^-3         (T0 = {:8.4g} K)".format(NV, T0Nh))
    print("   DC0 = {:12.6g} cm^-3/eV^1.5  (T0 = {:8.4g} K)".format(DC0, T0Ne))
    print("   DV0 = {:12.6g} cm^-3/eV^1.5  (T0 = {:8.4g} K)".format(DV0, T0Nh))
    print("   me* = {:8.4g} me".format(meeff))
    print("   mh* = {:8.4g} me".format(mheff))

    kBTe = kB * dos.T0 / e
    dEinteg = dos.nrange * kBTe
    E0h = dos.EV - dEinteg
    E1e = dos.EC + dEinteg

    e_original, dos_original, func = dos.get_dos()
    if e_original is not None and len(e_original) > 1:
        print("Effective density of states by integration:")
        dos.outrangeE = 0.0
#        elist, doslist = E_DOS_bt, DOS_bt
        elist, doslist = e_original, dos_original
        dos.set_dos(elist, doslist, EVBM_bt, ECBM_bt)
        NC2 = dos.NC_from_DOS(dos.T0, dos.EC, E1e, elist[1] - elist[0])
        NV2 = dos.NV_from_DOS(dos.T0, E0h, dos.EV, elist[1] - elist[0])
        meeff2 = NC2meff_FEA(NC2, dos.T0)
        mheff2 = NC2meff_FEA(NV2, dos.T0)
        DC02   = meff2DC0_FEA(meeff2, dos.T0)
        DV02   = meff2DC0_FEA(mheff2, dos.T0)
        print("   NC  = {:12.6g} cm^-3".format(NC2))
        
        
        print("   NV  = {:12.6g} cm^-3".format(NV2))
        print("   DC0 = {:12.6g} cm^-3/eV^1.5".format(DC02))
        print("   DV0 = {:12.6g} cm^-3/eV^1.5".format(DV02))
        print("   me* = {:8.4g} me".format(meeff2))
        print("   mh* = {:8.4g} me".format(mheff2))

    if E_vasp is not None and len(E_vasp) > 1:
        print("")
        print("Effective density of states by integrating VASP DOS (do not trust):")
        dos.outrangeE = 0.0
        elist, doslist = E_vasp, DOS_vasp
        dos.set_dos(elist, doslist, EVBM_bt, ECBM_bt)
        NC2 = dos.NC_from_DOS(dos.T0, dos.EC, E1e, elist[1] - elist[0])
        NV2 = dos.NV_from_DOS(dos.T0, E0h, dos.EV, elist[1] - elist[0])
        meeff2 = NC2meff_FEA(NC2, dos.T0)
        mheff2 = NC2meff_FEA(NV2, dos.T0)
        DC02   = meff2DC0_FEA(meeff2, dos.T0)
        DV02   = meff2DC0_FEA(mheff2, dos.T0)
        print("   NC  = {:12.6g} cm^-3".format(NC2))
        print("   NV  = {:12.6g} cm^-3".format(NV2))
        print("   DC0 = {:12.6g} cm^-3/eV^1.5".format(DC02))
        print("   DV0 = {:12.6g} cm^-3/eV^1.5".format(DV02))
        print("   me* = {:8.4g} me".format(meeff2))
        print("   mh* = {:8.4g} me".format(mheff2))

    if E_DOS_bt is not None and len(E_DOS_bt) > 1:
        print("")
        print("Effective density of states by integrating trace DOS (do not trust):")
        dos.outrangeE = 0.0
        elist, doslist = E_DOS_bt, DOS_bt
        dos.set_dos(elist, doslist, EVBM_bt, ECBM_bt)
        NC2 = dos.NC_from_DOS(dos.T0, dos.EC, E1e, elist[1] - elist[0])
        NV2 = dos.NV_from_DOS(dos.T0, E0h, dos.EV, elist[1] - elist[0])
        meeff2 = NC2meff_FEA(NC2, dos.T0)
        mheff2 = NC2meff_FEA(NV2, dos.T0)
        DC02   = meff2DC0_FEA(meeff2, dos.T0)
        DV02   = meff2DC0_FEA(mheff2, dos.T0)
        print("   NC  = {:12.6g} cm^-3".format(NC2))
        print("   NV  = {:12.6g} cm^-3".format(NV2))
        print("   DC0 = {:12.6g} cm^-3/eV^1.5".format(DC02))
        print("   DV0 = {:12.6g} cm^-3/eV^1.5".format(DV02))
        print("   me* = {:8.4g} me".format(meeff2))
        print("   mh* = {:8.4g} me".format(mheff2))

    if E_DOS_trbt is not None and len(E_DOS_trbt) > 1:
        print("")
        print("Effective density of states by integrating transdos DOS (trustable with very dens interpolation):")
        dos.outrangeE = None
        elist, doslist = E_DOS_trbt, DOS_trbt
        dos.set_dos(elist, doslist, EVBM_bt, ECBM_bt)
        NC2 = dos.NC_from_DOS(dos.T0, dos.EC, E1e, elist[1] - elist[0])
        NV2 = dos.NV_from_DOS(dos.T0, E0h, dos.EV, elist[1] - elist[0])
        meeff2 = NC2meff_FEA(NC2, dos.T0)
        mheff2 = NC2meff_FEA(NV2, dos.T0)
        DC02   = meff2DC0_FEA(meeff2, dos.T0)
        DV02   = meff2DC0_FEA(mheff2, dos.T0)
        print("   NC  = {:12.6g} cm^-3".format(NC2))
        print("   NV  = {:12.6g} cm^-3".format(NV2))
        print("   DC0 = {:12.6g} cm^-3/eV^1.5".format(DC02))
        print("   DV0 = {:12.6g} cm^-3/eV^1.5".format(DV02))
        print("   me* = {:8.4g} me".format(meeff2))
        print("   mh* = {:8.4g} me".format(mheff2))

    dos.set_dos(e_original, dos_original, EVBM_bt, ECBM_bt)

    DOS_FEA = []
    print("")
    print("Calculate free carrier approximation DOS")
#    print ("{:10}\t{:10}".format("E(eV)", "DOS,FEA(1/cm^3/eV)"))
    for i in range(nEF):
        EF = xEF[i]
        if EF < dos.EV:
            DOS_FEA.append(DV0 * sqrt(dos.EV - EF))
        elif EF > dos.EC:
            DOS_FEA.append(DC0 * sqrt(EF - dos.EC))
        else:
            DOS_FEA.append(0.0)
#        print ("{:10.4g}\t{:10.4g}".format(xEF[i], DOS_FEA[i]))

    xEFapprox = [EFmin + i * EFstep for i in range(nEF)]
    yNeapprox = [NC * exp(-(dos.EC - xEFapprox[i]) * e / kB / dos.T0) for i in range(nEF)]
    yNhapprox = [NV * exp(-(xEFapprox[i] - dos.EV) * e / kB / dos.T0) for i in range(nEF)]

    print("")
    print("Save fitting data to [{}]".format(outcsvfile))
    save_csv(outcsvfile, 
            ["EF(eV)", "Ne", "Nh", "ND+", "NA-", "RH", "T0(Ne)", "T0(Nh)", "Ns"], 
            [xEF, yNe, yNh, yNDp, yNAm, yRH, yT0Ne, yT0Nh, yNs])

    print("")
    print("Save parameters to [{}]".format(parameterfile))
    save_parameterfile(parameterfile)
    
#=============================
# Plot graphs
#=============================
    xrange = [max([EFmin, view_Emin]) - 10.0, min([EFmax, view_Emax]) + 10.0]
    d = []
    for i in range(len(dos.DOSlist)):
        if xrange[0] <= dos.Elist[i] <= xrange[1]:
            d.append(dos.DOSlist[i])
    yrange = [0.0, max(d) * 1.1]

    print("")
    print("plot")

    fig = plt.figure(figsize = figsize)

    axdos    = fig.add_subplot(2, 2, 1)
    axNinteg = axdos.twinx()
    axN   = fig.add_subplot(2, 2, 2)
    axRH  = fig.add_subplot(2, 2, 3)
    axT0  = fig.add_subplot(2, 2, 4)

    ins1 = axdos.plot(E_vasp,   DOS_vasp, label = 'DOS(VASP)',      color = 'black', linewidth = 0.5)
    if E_DOS_bt is not None:
        ins2 = axdos.plot(E_DOS_bt, DOS_bt,   label = 'DOS(trace)', color = 'red', linewidth = 0.5)
        ins1 += ins2
    if E_DOS_trbt is not None:
        ins2 = axdos.plot(E_DOS_trbt, DOS_trbt, label = 'DOS(transdos)', color = 'blue', linewidth = 0.5)
        ins1 += ins2
#        N_trbt     = inf_dosbt["N"]
    ins2 = axdos.plot(xEF,      DOS_FEA,  label = 'DOS$_{,FEA}$',   color = 'green', linewidth = 0.5, linestyle = 'dashed')
    ins3 = axdos.plot([dos.EF,  dos.EF],  [min(dos.DOSlist), max(dos.DOSlist)], label = '$E_F$',    color = 'red',  linestyle = 'dashed', linewidth = 0.5)
    ins4 = axdos.plot([dos.EF0, dos.EF0], [min(dos.DOSlist), max(dos.DOSlist)], label = '$E_{F0}$', color = 'pink', linestyle = 'dotted', linewidth = 0.3)
    ins5 = axdos.plot([dos.EV,  dos.EV],  [min(dos.DOSlist), max(dos.DOSlist)], label = '$E_V$',    color = 'blue', linestyle = 'dashed', linewidth = 0.5)
    ins6 = axdos.plot([dos.EC,  dos.EC],  [min(dos.DOSlist), max(dos.DOSlist)], label = '$E_C$',    color = 'blue', linestyle = 'dashed', linewidth = 0.5)
    ins1 += ins2 + ins3 + ins4 + ins5 + ins6
    axdos.set_xlim(xrange)
    axdos.set_ylim(yrange)
    axdos.set_xlabel("E (eV)")
    axdos.set_ylabel("DOS (states/cm$^3$)")
    if E_DOS_trbt is not None:
        ins2b = axNinteg.plot(E_DOS_trbt, N_trbt, label = 'N(transdos)(/unit cell)', linestyle = 'dashed', linewidth = 0.5, color = 'green')
        axNinteg.set_ylabel("N (cm$^{-3}$)")
        ins1 += ins2b
    axdos.legend(ins1, [l.get_label() for l in ins1], fontsize = legend_fontsize, loc = 'upper center') #loc = 'best')
#    axdos.legend()

    axN.plot(xEF, yNe,    label = '$N_e$',   color = 'black',  linewidth = 0.5)      #, marker = 'o')
    axN.plot(xEF, yNh,    label = '$N_h$',   color = 'red',    linewidth = 0.5)        #, marker = 'x')
    axN.plot(xEF, yNsabs, label = '|$N_s$|', color = 'purple', linewidth = 0.5, linestyle = 'dashed')     #, marker = '^')
#    axN.plot(xEF, yNDp,   label = '$N_D^+$', color = 'blue')    #, marker = 'x')
#    axN.plot(xEF, yNAm,   label = '$N_A^-$', color = 'green')   #, marker = 'x')
    axN.plot(xEFapprox, yNeapprox, label = 'Ne(Boltz)', color = 'gray',   linewidth = 0.5) #, marker = 'x')
    axN.plot(xEFapprox, yNhapprox, label = 'Nh(Boltz)', color = 'purple', linewidth = 0.5) #, marker = 'x')
    axN.plot([dos.EV, dos.EV], axN.get_ylim(), color = 'red', linestyle = 'dashed', linewidth = 0.5)
    axN.plot([dos.EC, dos.EC], axN.get_ylim(), color = 'red', linestyle = 'dashed', linewidth = 0.5)
    axN.set_yscale('log')
    axN.set_xlabel("EF (eV)")
    axN.set_ylabel("N (cm$^{-3}$)")
    axN.legend()

    axRH.plot(xEF, yRH, label = 'R$_H$', color = 'black', linewidth = 0.5, marker = 'o')
    axRH.set_xlabel("EF (eV)")
    axRH.set_ylabel("R$_H$ (C$^{-1}$cm$^{-3}$)")
    axRH.legend()

    axT0.plot(xEF, yT0Ne, label = 'T$_0$(Ne)', color = 'black')   #, marker = 'o')
    axT0.plot(xEF, yT0Nh, label = 'T$_0$(Nh)', color = 'red')     #, marker = 'x')
    axT0.plot([dos.EV, dos.EV], axT0.get_ylim(), color = 'red', linestyle = 'dashed', linewidth = 0.5)
    axT0.plot([dos.EC, dos.EC], axT0.get_ylim(), color = 'red', linestyle = 'dashed', linewidth = 0.5)
    axT0.set_ylim([0.0, 2.0*dos.T0])
    axT0.set_xlabel("EF (eV)")
    axT0.set_ylabel("T$_0$ (K)")
    axT0.legend()

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()
    plt.pause(0.1)

    print("")
    usage()
    print("")

    print("Press ENTER to exit>>", end = '')
    input()

    terminate("", usage = usage)

def cal_T():
    global mode
    global Work_dir
    global dEF0, nsmooth
    global E_raw, dos_raw, fdos, EDOSmin, EDOSmax, EDOSstep

    print("")
    print("===============================================================================")
    print(f"  Calculate T dependence of {mode}")
    print("===============================================================================")
    print("Temperature      : {} K".format(dos.T0))
    print("EF0 from EF(VASP): {:.4g} eV".format(dEF0))
    print("EF at {} K: {:.4g} eV".format(dos.T0, dos.EFatT0))
    print("tau0             : {} s".format(tau0))
    print("nsmooth          :", nsmooth)

    allinf     = read_files(infile)
    inf_inbt   = allinf.get("inf_intrans", None)
    inf_outbt  = allinf.get("inf_outputtrans", None)
    inf_trace  = allinf.get("inf_trace", None)
    outcsvfile = allinf.get("outcsvfile", None)
    outcarinf  = dos.vaspinf["OUTCAR"]
    cry        = dos.vaspinf["POSCAR"]
    doscarinf  = dos.vaspinf["DOSCAR"]
    ISPIN      = outcarinf["ISPIN"]
    EF_vasp0   = outcarinf["EF"]
    EF_vasp    = dos.EF # outcarinf["EF"]
    dos.EF0    = EF_vasp + dEF0
    E_vasp     = dos.Elist
    EV_vasp    = dos.EV
    EC_vasp    = dos.EC
    DOS_vasp   = dos.DOSlist
    print("  EF(VASP)(org): {} eV".format(EF_vasp0))
    print("  EF(VASP)     : {} eV".format(EF_vasp))
    print("  EF0          : {} eV".format(dos.EF0))
    print("  EV           : {} eV".format(dos.EV))
    print("  EC           : {} eV".format(dos.EC))
    print("  Vcell        : {} A^-3".format(dos.Vcell))
    print("  ISPIN: ", ISPIN)
    print("  EF(VASP) = {} eV".format(EF_vasp))

#    cry.PrintInf("cell")
    a, b, c, alpha, beta, gamm = cry.LatticeParameters()
    Vcell = cry.Volume()
    print("  cell: {:12.8f} {:12.8f} {:12.8f} A   {:10.6f} {:10.6f} {:10.6f}".format(a, b, c, alpha, beta, gamm))
    print("  volume: {:12.6f} A^-3".format(Vcell))

    EV    = dos.EV
    EC    = dos.EC
    Eg    = dos.Eg
    EHOMO = dos.EHOMO
    ELUMO = dos.ELUMO
    print("  Band edges: EV={:10.6f}  EC={:10.6f}  Eg={:10.6f}  EF(VASP)={:10.6f} eV"
        .format(EV, EC, Eg, EF_vasp))

    E_raw    = E_vasp
    dos_raw  = DOS_vasp
    Ne_raw   = doscarinf["Ne"]
    nDOS     = len(E_raw)
    Emin     = E_raw[0]
    Emax     = E_raw[nDOS-1]
    EDOSmin  = Emin
    EDOSmax  = Emax
    EDOSstep = E_raw[1] - E_raw[0]
    print("  DOS E range                     : {} - {}, {} eV step".format(EDOSmin, EDOSmax, EDOSstep))
#    print("{:8}\t{:12}".format('E(eV)', 'DOS'))
#    for i in range(nDOS):
#        print("{:8.4g}\t{:12.6g}".format(E_raw[i], dos_raw[i]))

    if inf_outbt is not None:
        print("")
        print("BoltzTraP2 info")
#        EVBM_bt = inf_outbt["EVBM"]
#        ECBM_bt = inf_outbt["ECBM"]
        print(".outputtrans exists but EV and EC are taken from VASP")
        EVBM_bt = EV_vasp
        ECBM_bt = EC_vasp
    else:
        print("")
        print(".outputtrans does not exist.")
        print(f"  EV and EC are taken from VASP")
        EVBM_bt = EV_vasp
        ECBM_bt = EC_vasp
    print("  EV           : {} eV".format(EVBM_bt))
    print("  EC           : {} eV".format(ECBM_bt))

    if inf_trace is None:
        E_DOS_bt     = None
        DOS_bt       = None
        DOS_bt_label = None
    else:
        E_DOS_bt     = inf_trace['xEF']
        DOS_bt       = inf_trace['dos'][0][0]
        DOS_bt_label = inf_trace['dos_label']
        DOS_source  = 'trace'
#    print("dos_bt", dos_bt)
#        dos.set_dos(E_DOS_bt, DOS_bt, EVBM = EVBM_bt, ECBM = ECBM_bt)

    inf_dosbt = allinf["inf_trancedos"]
    if inf_dosbt is None:
        nE         = 0
        E_DOS_trbt = None
        DOS_trbt   = None
        N_trbt     = None
    else:
        nE         = inf_dosbt["nE"]
        E_DOS_trbt = inf_dosbt["xE"]
        DOS_trbt   = inf_dosbt["dos"]
#        DOS_trbt   = simple_average(inf_dosbt["dos"], nsmooth)
#        DOS_trbt   = average_dos(E_DOS_trbt, inf_dosbt["dos"], nsmooth, dos.EV, dos.EC)
        N_trbt     = inf_dosbt["yNcm3"]
        DOS_source  = 'transdos'

    print("")
    print(f"Use {DOS_source} DOS to calculate NC and NV")
    if DOS_source == 'transdos':
        DOS_trbt   = average_dos(E_DOS_trbt, DOS_trbt, nsmooth, dos.EV, dos.EC)
    elif DOS_source == 'trace':
        DOS_bt   = average_dos(E_DOS_bt, DOS_bt, nsmooth, dos.EV, dos.EC)

    print("")
    print(f"Use {DOS_source} DOS to calculate NC and NV")
    if DOS_source == 'transdos':
        DOS_trbt = average_dos(E_DOS_trbt, DOS_trbt, nsmooth, EVBM_bt, ECBM_bt)
        dos.set_dos(E_DOS_trbt, DOS_trbt, EVBM = EVBM_bt, ECBM = ECBM_bt)
    elif DOS_source == 'trace':
        DOS_bt   = average_dos(E_DOS_bt, DOS_bt, nsmooth, EVBM_bt, ECBM_bt)
        dos.set_dos(E_DOS_bt, DOS_bt, EVBM = EVBM_bt, ECBM = ECBM_bt)

    condtenspath = allinf["condtenspath"]
    print("")
    print("Read condtens from {}".format(condtenspath))
    inf = BT.read_condtens(condtenspath, dos.Vcell, E0 = EF_vasp0, normalize_E= True, EFmin = EFmin, EFmax = EFmax, print_level = 0)
#    inf = BT.read_condtens(condtenspath, dos.Vcell, E0 = EF_vasp0, normalize_E= False, EFmin = EFmin, EFmax = EFmax, print_level = 0)
    if inf is None:
        terminate(f"Errorin cal_T(): Could not read [{condtenspath}]", usage = usage)

    header0      = inf['header']
    data_list    = inf['data_list']
    ncolumns     = inf['ncolumns']
    ndata        = inf['ndata']
    labels       = inf['labels']
    xT           = inf['xT']
    xEF          = inf['xEF']
#    for i in range(len(xEF)):
#        print("i=", i, xEF[i], xEF[i] / RyToeV, xEF[i] / RyToeV + EF_vasp0)
#    exit()        
    nT           = inf['nT']
    T_list       = inf['T_list']
    nEF          = inf['nEF']
    EF_list      = inf['EF_list']
    sigmalist    = inf['sigma']
    sigma_label  = inf['sigma_label']
    Slist        = inf['S']
    S_labels     = inf['S_label']
    kappalist    = inf['kappa']
    kappa_labels = inf['kappa_label']
    if mode == 'PF':
        plabels = ['PFxx', 'PFxy', 'PFxz', 'PFyx', 'PFyy', 'PFyz', 'PFzx', 'PFzy', 'PFzz']
    elif mode == 'ZT':
        plabels = ['ZTxx', 'ZTxy', 'ZTxz', 'ZTyx', 'ZTyy', 'ZTyz', 'ZTzx', 'ZTzy', 'ZTzz']
    else:
        plabels = inf[f'{mode}_label']
    nP = len(plabels)
    print("  ncolumns:", ncolumns)
    print("  ndata   :", ndata)
    print("  labels:", labels)
    print("  nT         :", nT)
    print("  nP         :", nP)
    print("  nEF        :", nEF)
    print("  T_list [K] :", T_list)
#    print("  EF_list [K]:", EF_list)

    dos.outrangeE = 0.0
    print("")
    print("*** Calculate EF(T)")
    print("  E(DOS) range: {} - {}, {} eV step".format(EDOSmin, EDOSmax, EDOSstep))
    fdos = interp1d(E_raw, dos_raw, kind = 'linear') #'cubic')

    Tmin = min(xT)
    Tmax = max(xT)
    kBTe = kB * Tmax / e
    dE = nrange * kBTe
    E0 = min([dos.EV, dos.EF0]) - dE
    E1 = max([dos.EC, dos.EF0]) + dE
    Estep = EF_list[1] - EF_list[0]
#    E0 = EDOSmin
#    E1 = EDOSmax
    print("  Integration E range: {:10.6g} - {:10.6g} eV at {} K".format(E0, E1, Tmax))

    Estep = EF_list[1] - EF_list[0]
    if abs(dos.EFatT0) < 5.0:
        dos.NeatT0 = dos.Ne(dos.T0, dos.EFatT0, E0, E1, Estep)
        Nh, Ne, NAm, NDp, RH, Ns = dos.cal_carrier_densities(dos.T0, dos.EFatT0, E0, dos.EV, dos.EC, E1, Estep)
        dos.dQatT0 = Ne + NAm - Nh - NDp
    else:
        print("  T0 = {} is larger than 5.0 so total Ne is calculated at 0 K and dQ(T0) is set to 0".format(dos.dQatT0))
        dos.NeatT0 = dos.Ne(0.0, dos.EV, E0, E1, Estep)
        dos.dQatT0 = 0.0
    print("  EF0    : {} eV".format(dos.EF0))
    print("  At {} K:".format(dos.T0))
    print("    EF: {} eV".format(dos.EFatT0))
    print("    Ne: {:12.6g} cm-3".format(dos.NeatT0))
    print("    dQ: {:12.6g} cm-3".format(dos.dQatT0))
    dos.EA = dos.EV + dos.dEA
    dos.ED = dos.EC + dos.dED
    print("  Doping:")
    print("    EV : {} eV".format(dos.EV))
    print("    EC : {} eV".format(dos.EC))
    print("    dEA: {} eV".format(dos.dEA))
    print("    EA : {} eV".format(dos.EA))
    print("    dED: {} eV".format(dos.dED))
    print("    ED : {} eV".format(dos.ED))
    EFmin_bs = min([dos.EFatT0, dos.EV]) - dos.dE_bisec
    EFmax_bs = max([dos.EFatT0, dos.EC]) + dos.dE_bisec
    print("  Bisection range: {:10.4g} - {:10.4g} eV".format(EFmin_bs, EFmax_bs))

# Temporary, dQatT0 is invalidated
#    dos.dQatT0 = 0.0
#    EF_s, dQ = dos.FindEF_semi(dos.T0, EFmin_bs, EFmax_bs, E0, dos.EV, dos.EC, E1, Estep, dQ0 = dos.dQatT0, print_level = 0)
#    exit()

    print("{:>8}\t{:>12}\t{:>12}\t{:>12}\t{:>12}\t{:>12}\t{:>12}\t{:>10}\t{:>12}"
                .format("T(K)", "EF,semi(eV)", "dQ", "EF,metal", "dNe", 'Nh,s', 'Ne,s', 'NA-,s', 'ND+,s'))
    yEF_s = []
    yEF_m = []
    EF_s = dos.EF0
    EF_m = EF_s
    for iT in range(len(xT)):
        T = xT[iT]

        EF_s, dQ = dos.FindEF_semi(T, EFmin_bs, EFmax_bs, E0, dos.EV, dos.EC, E1, Estep, dQ0 = dos.dQatT0, print_level = 0)
        Nh_s, Ne_s, NAm_s, NDp_s, RH_s, Ns_s = dos.cal_carrier_densities(T, EF_s, E0, dos.EV, dos.EC, E1, Estep)

# Temporary, EF_m is fixed constant
#        EF_m  = dos.EFatT0
# Examin Ne_m and Nh_m are satisfactory large to calculate EF,metal
        Ne_m = dos.Ne(T, dos.EFatT0, dos.EC, E1, Estep)
        Nh_m = dos.Nh(T, dos.EFatT0, E0, dos.EV, Estep)
#        print("Ne,m=", Ne_m, Nh_m)
        if Ne_m < 1.0e15 and Nh_m < 1.0e15:
            EF_m  = EF_s
            dNe_m = dQ
        else:
#            EF_m  = EF_s
#            dNe_m = dQ
            EF_m  = dos.FindEF_metal(T, EF_m, E0, E1, Estep, dos.NeatT0, 
                        hEF = 1.0e-7, dump = 1.0, maxiter = 100, ftol = 1.0, EFtol = 1.0e-3, print_level = 0)
            dNe_m = dos.Ne(T, EF_m, E0, E1, Estep) - dos.NeatT0

        yEF_s.append(EF_s)
        yEF_m.append(EF_m)
        print("{:8.3g}\t{:12.6g}\t{:12.6g}\t{:10.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}\t{:12.6g}"
                    .format(T, EF_s, dQ, EF_m, dNe_m, Nh_s, Ne_s, NAm_s, NDp_s))

    print("")
    print("*** Calculate temperature dependences of properties")
    _nsmooth = nsmooth
    if _nsmooth > 1:
        _nsmooth = int(nsmooth / 2.0) * 2 + 1
    EFstep = xEF[1] - xEF[0]
    prop3  = np.empty([nT, nP, nEF], dtype = float)
    prop_s = np.empty([nP, nT], dtype = float)
    prop_m = np.empty([nP, nT], dtype = float)
    for iP in range(nP):
        for iT in range(nT):
            for iEF in range(nEF):
                if mode == 'S':
                    prop3[iT][iP][iEF] = Slist[iT][iP][iEF]
                elif mode == 'sigma':
                    prop3[iT][iP][iEF] = sigmalist[iT][iP][iEF] * tau0
                elif mode == 'kappa':
                    prop3[iT][iP][iEF] = kappalist[iT][iP][iEF] * tau0
                elif mode == 'PF':
                    prop3[iT][iP][iEF] = sigmalist[iT][iP][iEF] * tau0 * Slist[iT][iP][iEF]**2
                elif mode == 'ZT':
                    prop3[iT][iP][iEF] = sigmalist[iT][iP][iEF] * tau0 * Slist[iT][iP][iEF]**2 \
                                         / (kappalist[iT][iP][iEF] * tau0 + kappa_latt) * xT[iT]
#                print("i=", iP, iT, iEF, prop3[iT][iP][iEF])

    print("nsmooth:", _nsmooth)
    print("EF step: {} eV".format(EFstep))
    print("{:>6}\t{:>8}\t{:>8}\t".format('T', 'EF,semi', 'EF,metal'), end = '')
    for iP in range(nP):
        print("{:^10} {:^10} ".format(plabels[iP]+'_s', plabels[iP]+'_m'), end = '')
    print("")
    for iT in range(len(xT)):
        T    = xT[iT]
        EF_s = yEF_s[iT]
        EF_m = yEF_m[iT]
        print("{:6.3g}\t{:8.3g}\t{:8.3g}\t".format(T, EF_s, EF_m), end = '')
        for iP in range(nP):
            func = interp1d(xEF, prop3[iT][iP], kind = 'linear') #'cubic')
#            func = interp1d(xEF, ydata[iT][iP], kind = 'linear') #'cubic')

            if nsmooth <= 1:
                p_s = func(EF_s)
                p_m = func(EF_m)
                prop_s[iP][iT] = p_s
                prop_m[iP][iT] = p_m
            else:
                p_s = 0.0
                p_m = 0.0
                n = 0
                for i in range(-int(_nsmooth/2), int(_nsmooth/2) + 1):
                    p_s += func(EF_s + i * EFstep)
                    p_m += func(EF_m + i * EFstep)
                    n += 1
                prop_s[iP][iT] = p_s / n
                prop_m[iP][iT] = p_m / n

            print("{:10.3g} {:10.3g} ".format(p_s, p_m), end = '')
        print("")

    fname = "{}-semi-T.csv".format(mode)
    print("  Save to [{}]".format(fname))
    fp = tkFile(fname, 'w')
    if not fp:
        terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
    fp.Write("T(K),EF(eV)," + ",".join(plabels) + "\n")
    for iT in range(nT):
        v_labels = ["{:g}".format(prop_s[iP][iT]) for iP in range(nP)]
        fp.Write("{},{},".format(xT[iT], yEF_s[iT]) + ",".join(v_labels) + "\n")
    fp.Close()

    fname = "{}-metal-T.csv".format(mode)
    print("  Save to [{}]".format(fname))
    fp = tkFile(fname, 'w')
    if not fp:
        terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
    fp.Write("T(K),EF(eV)," + ",".join(plabels) + "\n")
    for iT in range(nT):
        v_labels = ["{:g}".format(prop_m[iP][iT]) for iP in range(nP)]
        fp.Write("{},{},".format(xT[iT], yEF_m[iT]) + ",".join(v_labels) + "\n")
    fp.Close()

    print("")
    print("Save parameters to [{}]".format(parameterfile))
    save_parameterfile(parameterfile)


#=============================
# グラフの表示
#=============================
    xrange = [max([Emin, view_Emin]), min([Emax, view_Emax])]
    Trange = [min(xT), max(xT)]
    yrange = [0.0, max(dos_raw) * 1.1]

    print("")
    print("plot")
    
    fig = plt.figure(figsize = figsize)

    axdos  = fig.add_subplot(3, 4, 1)
    axN    = axdos.twinx()
    axEF   = fig.add_subplot(3, 4, 5)
    axprop = fig.add_subplot(3, 4, 9)
    axP    = []
    axP.append(fig.add_subplot(3, 4,  2))
    axP.append(fig.add_subplot(3, 4,  3))
    axP.append(fig.add_subplot(3, 4,  4))
    axP.append(fig.add_subplot(3, 4,  6))
    axP.append(fig.add_subplot(3, 4,  7))
    axP.append(fig.add_subplot(3, 4,  8))
    axP.append(fig.add_subplot(3, 4, 10))
    axP.append(fig.add_subplot(3, 4, 11))
    axP.append(fig.add_subplot(3, 4, 12))
#    axPT = []
#    for iP in range(9):
#        axPT.append(axP[iP].twiny())

    axdos.set_title(infile)

    if DOS_source == 'transdos':
        axdos.plot(E_DOS_trbt,  DOS_trbt, label = 'DOS(trancedos)', color = 'black', linewidth =0.5)
    elif DOS_source == 'trace':
        axdos.plot(E_DOS_bt,  DOS_bt, label = 'DOS(trace)', color = 'black', linewidth =0.5)
    else:
        axdos.plot(E_raw,  dos_raw, label = 'DOS', color = 'black', linewidth =0.5)
    axdos.plot([dos.EV, dos.EV],   yrange, label = '$E_V$',      color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
    axdos.plot([dos.EC, dos.EC],   yrange, label = '$E_C$',      color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
    axdos.plot([EF_vasp, EF_vasp], yrange, label = '$E_{VASP}$', color = 'cyan', linestyle = 'dashed', linewidth = 0.5)
    axdos.plot([dos.EF0, dos.EF0], yrange, label = '$E_{F0}$',   color = 'green', linestyle = 'dashed', linewidth = 0.5)
    axdos.set_xlim([xrange[0] - 10.0, xrange[1] + 10.0])
    axdos.set_ylim([0.0, None])
    axdos.set_xlabel("$E - E_F$ (eV)")
    axdos.set_ylabel("DOS (states/cm$^3$)")
    axdos.legend(fontsize = legend_fontsize)

    yrange = [0.0, max(Ne_raw) * 1.1]
    axN.plot(E_raw,  Ne_raw, label = '$N_e$', color = 'red', linewidth =0.5)
    axN.plot([dos.EV, dos.EV],   yrange, label = '$E_V$',      color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
    axN.plot([dos.EC, dos.EC],   yrange, label = '$E_C$',      color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
    axN.plot([EF_vasp, EF_vasp], yrange, label = '$E_{VASP}$', color = 'cyan', linestyle = 'dashed', linewidth = 0.5)
    axN.plot([dos.EF0, dos.EF0], yrange, label = '$E_{F0}$',   color = 'green', linestyle = 'dashed', linewidth = 0.5)
    axN.set_xlim(xrange)
    axN.set_ylim([0.0, None])
#    axN.set_xlabel("$E - E_F$ (eV)")
    axN.set_ylabel("$N_e$ (/cm$^3$)")
#    axN.legend(fontsize = legend_fontsize)

    axEF.plot(xT, yEF_s, label = '$E_F$$_{,semi}$',  color = 'black', linewidth =0.5, marker = 'o', markersize = 6.0, markerfacecolor = 'w')
    axEF.plot(xT, yEF_m, label = '$E_F$$_{,metal}$', color = 'blue',  linewidth =0.5, marker = 'x', markersize = 6.0, markerfacecolor = 'w')
    axEF.plot(Trange, [dos.EV, dos.EV], label = '$E_V$', color = 'blue',   linestyle = 'dashed', linewidth = 0.5)
    axEF.plot(Trange, [dos.EC, dos.EC], label = '$E_C$', color = 'blue',   linestyle = 'dashed', linewidth = 0.5)
    axEF.plot(Trange, [dos.EA, dos.EA], label = '$E_V$', color = 'green',  linestyle = 'dashed', linewidth = 0.5)
    axEF.plot(Trange, [dos.ED, dos.ED], label = '$E_V$', color = 'purple', linestyle = 'dashed', linewidth = 0.5)
    axEF.set_xlabel("$T$ (K)")
    axEF.set_ylabel("$E_F$ (eV)")
    axEF.legend(fontsize = legend_fontsize)

    for iP in range(nP):
        axprop.plot(xT, prop_s[iP], label = plabels[iP]+'_s', linewidth =0.5, marker = 'o', markersize = 6.0, markerfacecolor = 'w')
        axprop.plot(xT, prop_m[iP], label = plabels[iP]+'_m', linewidth =0.5, marker = 'x', markersize = 6.0, markerfacecolor = 'w')
    axprop.set_xlabel("$T$ (K)")
    axprop.set_ylabel(plabels[0].replace('xx', '$_{ij}$'))
#    axprop.legend(fontsize = legend_fontsize)

    nTstep = int(nT / nT_view + 0.99999)
    print("plot temperature variations of properties")
    print("  nT_view:", nT_view)
    print("  nTstep :", nTstep)
    for iT in range(0, nT, nTstep):
        print("    T={:8.3g} K".format(xT[iT]))
    for iP in range(0, len(axP)):
        offset = max(prop3[0][iP]) - min(prop3[0][iP])
#        offset = max(ydata[0][iP]) - min(ydata[0][iP])
        print("  iP, offset=", iP, offset)
        for iT in range(0, nT, nTstep):
            off = offset * iT / nTstep
            axP[iP].plot(xEF,  prop3[iT][iP] + off, label = "{} K".format(xT[iT]), linewidth =0.5)
#            axP[iP].plot(xEF,  ydata[iT][iP] + off, label = "{} K".format(xT[iT]), linewidth =0.5)
            axP[iP].plot(xrange, [off, off], linestyle = 'dotted', color = 'red', linewidth =0.5)
            axP[iP].plot([yEF_s[iT]], [prop_s[iP][iT] + off], linestyle = '', marker = 'o', markersize = 6.0, markeredgewidth =  0.5, markerfacecolor = 'none')
            axP[iP].plot([yEF_m[iT]], [prop_m[iP][iT] + off], linestyle = '', marker = 'x', markersize = 6.0, markeredgewidth =  0.5, markerfacecolor = 'none')
    for iP in range(0, len(axP)):
        yrange = axP[iP].get_ylim()
        axP[iP].plot([dos.EV, dos.EV],   yrange, label = '$E_V$',        color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
        axP[iP].plot([dos.EC, dos.EC],   yrange, label = '$E_C$',        color = 'blue',  linestyle = 'dashed', linewidth = 0.5)
        axP[iP].plot([EF_vasp, EF_vasp], yrange, label = '$E_{F,VASP}$', color = 'cyan',  linestyle = 'dashed', linewidth = 0.5)
        axP[iP].plot([dos.EF0, dos.EF0], yrange, label = '$E_{F0}$',     color = 'green',  linestyle = 'dashed', linewidth = 0.5)
        axP[iP].set_xlim(xrange)
        axP[iP].set_xlabel("$E - E_F$ (eV)")
        axP[iP].set_ylabel(plabels[iP])
        axP[iP].legend(fontsize = legend_fontsize)

#        axPT[iP].plot(xT, prop[iP], color = 'red', linewidth =0.5)
#        axPT[iP].set_xlabel("$E - E_F$ (eV)")
#        axPT[iP].set_ylabel("$N_e$ (/cm$^3$)")

# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    print("")
    print("Close graph window to terminate")
    plt.show()

    """
    plt.pause(0.1)
    print("")
    print("Press ENTER to exit>>", end = '')
    input()
    """

    terminate("", usage = usage)

def convert():
    global mode
    global Work_dir
    global dEF0, kappa_latt

    kBTe = kB * dos.T0 / e
    dE = nrange * kBTe

    print("")
    print("===============================================================================")
    print(f"  Calculate T dependence of {mode}")
    print("===============================================================================")
    print("conv_mode:", conv_mode)
    print("Temperature: {} K".format(dos.T0))
    print("EF(0K) from EF(VASP): {:14.6g} eV".format(dEF0))
    print("tau0      : {} s".format(tau0))
    print("kappa_latt: {} W/m/K".format(kappa_latt))

    allinf     = read_files(infile)
    inf_inbt   = allinf.get("inf_intrans", None)
    inf_outbt  = allinf.get("inf_outputtrans", None)
    inf_trace  = allinf.get("inf_trace", None)
    outcsvfile = allinf.get("outcsvfile", None)
    outcarinf  = dos.vaspinf["OUTCAR"]
    cry        = dos.vaspinf["POSCAR"]
    doscarinf  = dos.vaspinf["DOSCAR"]
    ISPIN      = outcarinf["ISPIN"]
    EF_vasp0   = outcarinf["EF"]
    EF_vasp    = dos.EF # outcarinf["EF"]
    print("")
    print("Information in OUTCAR:")
    print("  ISPIN: ", ISPIN)
    print("  EF(VASP) = {} eV".format(EF_vasp))
#    cry.PrintInf("cell")
    a, b, c, alpha, beta, gamm = cry.LatticeParameters()
    Vcell = cry.Volume()
    print("  cell: {:12.8f} {:12.8f} {:12.8f} A   {:10.6f} {:10.6f} {:10.6f}".format(a, b, c, alpha, beta, gamm))
    print("  volume: {:12.6f} A^-3".format(Vcell))

    condtenspath = allinf["condtenspath"]
    print("")
    print("Read condtens from {}".format(condtenspath))
    inf = BT.read_condtens(condtenspath, dos.Vcell, E0 = EF_vasp0, normalize_E= True, EFmin = EFmin, EFmax = EFmax, print_level = 0)
    header0      = inf['header']
    data_list    = inf['data_list']
    ncolumns     = inf['ncolumns']
    ndata        = inf['ndata']
    labels       = BT.get_labels(latex_style = 0)
    T_list       = inf['T_list']
    EF_list      = inf['EF_list']
    print("  ncolumns:", ncolumns)
    print("  ndata   :", ndata)
    print("  labels:", labels)

    halltenspath = allinf["halltenspath"]
    print("")
    print("Read halltens from {}".format(halltenspath))
    infhall = BT.read_halltens(halltenspath, dos.Vcell, E0 = EF_vasp0, normalize_E= True, EFmin = EFmin, EFmax = EFmax, print_level = 0)
    RH       = infhall['RH']
    RH_label = infhall['RH_label']
    print("  RH label:", RH_label)
 
    print("")
    print("===========================================")
    print(" Save property tensors to separated files")
    print("===========================================")
    if conv_mode == 'all':
        xEF, xT, yNcm3, ydata, plabels = BT.extract_data(labels, T_list, data_list, Vcell, 'all')
        nT = len(xT)
        nP = len(ydata[0])
        nT = len(xT)
        nP = len(ydata[0])
        plabels = plabels \
                     + ['ktotxx', 'ktotxy', 'ktotxz', 'ktotyx', 'ktotyy', 'ktotyz', 'ktotzx', 'ktotzy', 'ktotzz'] \
                     + ['PFxx', 'PFxy', 'PFxz', 'PFyx', 'PFyy', 'PFyz', 'PFzx', 'PFzy', 'PFzz'] \
                     + ['ZTxx', 'ZTxy', 'ZTxz', 'ZTyx', 'ZTyy', 'ZTyz', 'ZTzx', 'ZTzy', 'ZTzz']\
                     + RH_label \
                     + ['Nxyz', 'Nxzy', 'Nyxz', 'Nyzx', 'Nzxy', 'Nzyx'] 
        print("labels:", plabels)
        for iT in range(len(xT)):
            fname = "{}-{}K.csv".format('all', xT[iT])
            print("  Save to [{}]".format(fname))
            fp = tkFile(fname, 'w')
            if not fp:
                terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
            fp.Write("EF(eV)," + ",".join(plabels) + "\n")
            for iEF in range(len(ydata[0][0])):
                sigma     = []
                S         = []
                kappa     = []
                kappa_tot = []
                PF        = []
                ZT        = []
                _RH       = []
                N         = []
                for i in range(9):
                    sigma.append    (ydata[iT][i][iEF] * tau0)
                    S.append        (ydata[iT][i + 9][iEF])
                    kappa.append    (ydata[iT][i + 18][iEF] * tau0)
                    kappa_tot.append(kappa[i] + kappa_latt)
                    PF.append       (sigma[i] * S[i] * S[i])
                    ZT.append(PF[i] / (kappa[i] + kappa_latt) * xT[iT])
                for i in range(27):
                    _RH.append(RH[iT][i][iEF])
                for idx in [5, 7, 11, 15, 19, 21]: 
                    if RH[iT][idx][iEF] == 0.0:
                        N.append(0.0)
                    else:
                        N.append(1.0 / e / RH[iT][idx][iEF])
                v_list = sigma + S + kappa + kappa_tot + PF + ZT + _RH + N
                v_labels = ["{:g}".format(v_list[iP]) for iP in range(len(v_list))]
                fp.Write("{},".format(xEF[iEF]) + ",".join(v_labels) + "\n")
            fp.Close()
    else:
        for md in ('S', 'sigma', 'kappa'):
            print("")
            xEF, xT, yNcm3, ydata, plabels = BT.extract_data(labels, T_list, data_list, Vcell, md)
            nT = len(xT)
            nP = len(ydata[0])
            nproperties = len(ydata)
            print("labels:", plabels)
            if conv_mode == 'T':
                for iP in range(len(plabels)):
                    fname = "{}.csv".format(plabels[iP])
                    print("  Save to [{}]".format(fname))
                    fp = tkFile(fname, 'w')
                    if not fp:
                        terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
                    T_labels = ["{}K".format(v) for v in T_list]
                    fp.Write("EF(eV)," + ",".join(T_labels) + "\n")
                    for iEF in range(len(ydata[0][0])):
                        if md == 'S':
                            v_labels = ["{:g}".format(ydata[iT][iP][iEF]) for iT in range(nT)]
                        else:
                            v_labels = ["{:g}".format(ydata[iT][iP][iEF] * tau0) for iT in range(nT)]
                        fp.Write("{},".format(xEF[iEF]) + ",".join(v_labels) + "\n")
                    fp.Close()
            elif conv_mode == 'tensor':
                for iT in range(len(xT)):
                    fname = "{}-{}K.csv".format(md, xT[iT])
                    print("  Save to [{}]".format(fname))
                    fp = tkFile(fname, 'w')
                    if not fp:
                        terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
                    fp.Write("EF(eV)," + ",".join(plabels) + "\n")
                    for iEF in range(len(ydata[0][0])):
                        if md == 'S':
                            v_labels = ["{:g}".format(ydata[iT][iP][iEF]) for iP in range(nP)]
                        else:
                            v_labels = ["{:g}".format(ydata[iT][iP][iEF] * tau0) for iP in range(nP)]
                        fp.Write("{},".format(xEF[iEF]) + ",".join(v_labels) + "\n")
                    fp.Close()
            else:
                terminate("Error in convert(): Invalid conv_mode [{}]".format(conv_mode), usage = usage)

    fname = "Ne.csv"
    print("")
    print("  Save to [{}]".format(fname))
    fp = tkFile(fname, 'w')
    if not fp:
       terminate("Error in convert(): Can not write to [{}]".format(fname), usage = usage)
    T_labels = ["{}K".format(v) for v in T_list]
    fp.Write("EF(eV)," + ",".join(T_labels) + "\n")
    for iEF in range(len(ydata[0][0])):
         v_labels = ["{:g}".format(yNcm3[iT][iEF]) for iT in range(nT)]
         fp.Write("{},".format(xEF[iEF]) + ",".join(v_labels) + "\n")
    fp.Close()

    print("")
    print("Save parameters to [{}]".format(parameterfile))
    save_parameterfile(parameterfile)

    terminate("", usage = usage)


def main():
    updatevars()

    print("")
    print("============================================================================================================")
    print("  Calculate temperature dependences of transport properties based on BoltzTraP2 output")
    print("============================================================================================================")

    print("")
    print("infile               : {}".format(infile))
    print("parameter file       : {}".format(parameterfile))
    print("parameter backup file: {}".format(parameterbkfile))
    print("mode: ", mode)
    print("")

    print("")
    print("Read [{}]".format(parameterfile))
    read_parameters(parameterfile)

# check args again to use given parameters
    updatevars()

    print("")
    print("Parameters")
    print_parameters() #varname, varunit, dos.parameter_list(), optid)

    if mode == 'NC':
        cal_NC()
    elif mode == 'S' or mode == 'sigma' or mode == 'kappa' or mode == 'PF' or mode == 'ZT':
        cal_T()
    elif mode == 'conv':
        convert()
    else:
        terminate("Error in main(): Invalide mode [{}]".format(mode), usage = usage)


if __name__ == "__main__":
    main()
