import numpy as np
from math import sqrt, pow, exp
import os
import sys
import csv
import pandas as pd
import openpyxl
from matplotlib import pyplot as plt
from matplotlib import gridspec


from tklib.tkutils import terminate, pfloat, pint, getarg, getintarg, getfloatarg
from tklib.tkapplication import tkApplication
from tklib.tkvariousdata import tkVariousData


#===================================
# physical constants
#===================================
pi   = 3.14159265358979323846
e    = 1.60218e-19      # C";
kB   = 1.380658e-23     # JK<sup>-1</sup>";

kBe = kB / e             #8.6173303e-5      # kB/e


# Global variables
infile  = "pisarenko-plot-all-data-STO.xlsx"
outfile = None

T_label = 0
S_label = 1
N_label = 3


#=============================
# Graph configuration
#=============================
fig = None
figsize  = (6, 6)
fontsize = 18
legend_fontsize = 12

app = None

#===================================
# Treat arguments
#===================================
infile  = getarg(1, defval = infile)
T_label = getarg(2, defval = T_label)
S_label = getarg(3, defval = S_label)
N_label = getarg(4, defval = N_label)

header, ext = os.path.splitext(infile)
filebody    = os.path.basename(header)
outfile     = f'{header}-meff-out.xlsx'


#===================================
# Other functions
#===================================
def usage(app):
    argv = sys.argv
    print("")
    print("Usage: python {} infile T_label S_label(V/K) N_label(cm^-3)".format(argv[0]))
    print("   ex: python {} {} {} {} {}".format(argv[0], 
                    infile, T_label, S_label, N_label))

def savecsv(outfile, header, datalist):
    try: 
        print("Write to [{}]".format(outfile))
        f = open(outfile, 'w')
    except:
#    except IOError:
        print("line 98 Error: Can not write to [{}]".format(outfile))
    else:
        fout = csv.writer(f, lineterminator='\n')
        fout.writerow(header)
#        fout.writerows(data)
        for i in range(0, len(datalist[0])):
            a = []
            for j in range(len(datalist)):
                a.append(datalist[j][i])
            fout.writerow(a)
        f.close()

def read_file(fname):
    print("")

    datafile = tkVariousData(infile)
    labels, datalist = datafile.Read_minimum_matrix(close_fp = True, usage = usage)

    return datafile, labels, datalist

#Calculation method
def cal_ke(sigma, L, T):
    return sigma * L * T

def Fr(EF, r):
    result = integrate.quad(lambda E: E**r / (1 + np.exp(E - EF)), 0, np.inf)
    #print("felmi_integral = {}".format(result))
    fel = float(result[0])
    #print(fel)
    return fel

def S_comparison(T, S, sigma, EF):
    diff = S - kBe * (2.0 * Fr(EF, r = 1.0) / Fr(EF, r = 0.0) - EF)
    #print("diff = {}".format(diff))
    return diff

def Lorenz(EF):
    answer = kBe**2 * (3.0 * Fr(EF, r = 2.0) / Fr(EF, r = 0.0) - (2.0 * Fr(EF, r = 1.0) / Fr(EF, r = 0.0))**2)
    return answer

def bisection(T, S, sigma, max, min, err):
    xEF = (max + min) / 2.0
    ymin = S_comparison(T, S, sigma, min)
    ymax = S_comparison(T, S, sigma, max)
    yh   = S_comparison(T, S, sigma, xEF)
    print(f"{xEF=} {ymin=} {ymax=} {yh=} {min=} {max=}")

    if ymin * ymax > 0.0:
        print("line Error: ymin * ymax > 0.0")
        app.terminate(pause = True)

    for i in range(nmaxiter):
        print(f"xEF={xEF:8.3f} d=({ymin:8.4g} {yh:8.4g} {ymax:8.4g}) min={min:8.3f} max={max:8.3f}")

        if abs(yh) < eps:
#        if abs(yh) < eps or abs(max - min) < eps:
            return xEF
        
        if ymin * yh > 0.0:
            min = xEF
        else:
            max = xEF

        xEF = (max + min) / 2.0
        ymin = S_comparison(T, S, sigma, min)
        ymax = S_comparison(T, S, sigma, max)
        yh   = S_comparison(T, S, sigma, xEF)
    else:
        print(f"Did not converge in {nmaxiter} iterations.")
        exit()

def cal_Ne(T, xEF, NC):
    F05 = Fr(xEF, r = 0.5)
    N = NC * kpi05 * F05
    return N

def main():
    global app
    
    app = tkApplication()
    logfile = app.replace_path(infile)
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    print("")
    print("input file :", infile)
    print("output file:", outfile)
    print("T_label    :", T_label)
    print("S_label    :", S_label)
    print("N_label    :", N_label)
    print("")
    
    if T_label is str and '***' in T_label:
        app.terminate("Error: Choose T", usage = usage, pause = True)
    if S_label is str and '***' in S_label:
        app.terminate("Error: Choose S", usage = usage, pause = True)
    if N_label is str and '***' in N_label:
        app.terminate("Error: Choose N", usage = usage, pause = True)

    if S_label is str and ('uK' in S_label or 'microK' in S_label):
        app.terminate(f"Error: The unit of S must be 'V/K'. The given label is [{S_label}]", usage = usage, pause = True)

    print("Read data from [{}]".format(infile))
    datafile, labels, datalist = read_file(infile)
    T_idx = datafile.FindLabelIndex(T_label, flag = 'i')
    S_idx = datafile.FindLabelIndex(S_label, flag = 'i')
    N_idx = datafile.FindLabelIndex(N_label, flag = 'i')
    print("T index: ", T_idx)
    print("S index: ", S_idx)
    print("N index: ", N_idx)
    Tlabel, Tlist         = datafile.FindDataArray(T_label, flag = 'i')
    Slabel, Slist         = datafile.FindDataArray(S_label, flag = 'i')         # V/K
    Nlabel, Nlist         = datafile.FindDataArray(N_label, flag = 'i')         # cm-3

    print("")
    print(f"{'T(K)':8}  {'S(V/K)':10}  {'N(cm^-3)':10}")
    for i in range(len(Tlist)):
        print(f"{Tlist[i]:8.3g}  {Slist[i]:10.4g}  {Nlist[i]:10.4g}")

    meff_list = []
    for ic in range(len(Tlist)):
        T = Tlist[ic]
        S = Slist[ic]
        N = Nlist[ic]
        print(f"{ic:3d}: {T=} K  {S=} V/K  {N=} cm^-3")

        SkBe = S / kBe
        a = SkBe - 1.0 / SkBe
        expSkBe2 = exp(SkBe - 2.0)
        if expSkBe2 - 0.17 >= 0.0:
            A1 = 3.0 * pow(expSkBe2 - 0.17, 2.0/3.0)
            A2 = 1.0 + exp(-5.0 * a)
            B2 = 1.0 + exp(5.0 * a)
            meff = 0.924 * (300.0 / T) * pow(N / 1.0e20, 2.0/3.0) * (A1 / A2 + SkBe / B2)
        else:
            meff = 0.0

        print(f"T    = {T} K")
        print(f"S    = {S*1.0e6} uV/K")
        print(f"N    = {N} cm^-3")
        print(f"meff = {meff} me")

        meff_list.append(meff)

    df = pd.DataFrame(np.array([Tlist,  Nlist,      Slist,    meff_list]).T,
                     columns = ["T(K)", "N(cm^-3)", "S(V/K)", "meff(me)"])
    print("")
    print("Save meff data to [{}]".format(outfile))
    df.to_excel(outfile, index = False, header = True)


#=============================
# グラフの表示
#=============================
    print("")

    fig = plt.figure(figsize = figsize)
    spec = gridspec.GridSpec(ncols = 1, nrows = 2, height_ratios=[1, 1])
    ax1 = fig.add_subplot(spec[0])
    ax2 = fig.add_subplot(spec[1], sharex = ax1)

    ax1.tick_params(labelbottom = False, labelsize = fontsize)
    ax2.tick_params(labelsize = fontsize)

    ax1.plot(Nlist, Slist, label = '$S$ (V/K)', linestyle = '', marker = 'o')
#    ax1.set_xlabel("$N_{Hall}$ (cm$^{-3}$)", fontsize = fontsize)
    ax1.set_ylabel("$S$ (V/K)", fontsize = fontsize)
    ax1.set_xscale('log')

    ax2.plot(Nlist, meff_list, label = '$m_{eff}$', linestyle = '', marker = 's')
    ax2.set_xlabel("$N_{Hall}$ (cm$^{-3}$)", fontsize = fontsize)
    ax2.set_ylabel("$m_{eff}$", fontsize = fontsize)
    ax2.set_xscale('log')

    xlim = ax1.get_xlim()
    ax1.plot(xlim, [0.0, 0.0], linestyle = 'dashed', linewidth = 0.5, color = 'red')
    ax2.plot(xlim, [0.0, 0.0], linestyle = 'dashed', linewidth = 0.5, color = 'red')

#    ax1.legend(fontsize = legend_fontsize)
#    ax1b.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)

    app.terminate("", usage = usage, pause = True)


if (__name__ == '__main__'):
    main()

